/*
 * ToAmiga.c - A Music Module Converter  Version 1.04
 * Copyright (c) 1992 by L.Georgiev
 */

/*****************************************************************************
* Converts Scream Tracker modules to Amiga SoundTracker/NoiseTracker modules *
******************************************************************************/
/*
 * Changes since version 1.00 released:
 *  - now effects are processed even if there isn't a note info.
 *  - patterns present in sequence but not existant are ignored.
 *  - warnings when early EOF reached ( sample data truncated ).
 *  - appends identification string after last sample data byte.
 * (17-18.IV.1992)
 */

#include <io.h>
#include <conio.h>
#include <stdio.h>
#include <fcntl.h>          /* O_ constant definitions */
#include <sys\types.h>
#include <sys\stat.h>       /* S_ constant definitions */
#include <stdlib.h>
#include <malloc.h>
#include <errno.h>

#include "toamiga.h"


main( int argc, char *argv[] )
{
    if( argc == 3 )
        if( read_conv_write( argv[1], argv[2] ) )
            perror( "ToAmiga" );
	else
	    errno = 0;
    else        {
	printf
      (  "ToAmiga Music Module Converter  Ver %d.%02d\n", VER/100, VER%100);
printf(  "Copyright (c) MCMXCII  .eopue, BAPHA\n\n" );

printf(  "Usage: ToAmiga <source.stm> <target.mod>\n" );

printf("\nConverts a Scream Tracker module to an Amiga SoundTracker/NoiseTracker module.");
printf("\nSource and/or target may include drive and/or path.  No wildcards are allowed.");
printf("\nYou may freely use and distribute this program in its unaltered form." );

printf("\n\n                 *** Was free,  Is free,  Always free ***" );
	errno = EINVAL;
    }                                           /* Even upon abnormal exit  */
    return( exit( errno ));                     /* flush buf's & close files*/
}                                               /* MSDOS ERRORLEVEL = errno */

/*****************************************************************************
*                              Sub-main function                             *
******************************************************************************/

/* For each module item reads source, calls convert function and writes target
 * (both specified by path). Dynamically allocates memory for the file buffer.
 * Prompts before overwriting existing file.
 * Returns 0 if successful, or an error number if unsuccessful.
 */
int read_conv_write( char *source, char *target )
{
    register	word	    count,  count1, i, j;
    register	word	    *buf,   *buf1;
		int	    hsource, htarget, ch;
    struct	hdr_info    inf;
		char	    *eff;
		char	    id[] = " ToA104";

/******* Open source file and create target, overwriting if necessary *******/

    if( (hsource = open( source, O_BINARY | O_RDONLY )) == - 1 )
        return errno;
    htarget = open( target, O_BINARY | O_WRONLY | O_CREAT | O_EXCL,
                            S_IREAD | S_IWRITE );
    if( errno == EEXIST )
    {
	printf( "\"%s\" exists. Overwrite? ", target );
        ch = getch();
	if( ch == 'y' ||  ch == 'Y')        {
	    printf("yes\n");
	    errno = 0;
	    htarget = open( target, O_BINARY | O_WRONLY | O_CREAT | O_TRUNC,
                                    S_IREAD | S_IWRITE );
        }
        else
	    printf("no\n");
    }
    if( htarget == -1 )
	return errno;

/************************* Convert sample attributes ************************/

    count = sizeof( struct stm_header ) +
	SAMPLES * sizeof( struct stm_instr ) + PATTERNS;
    if( (buf = (word *)calloc( (size_t)count, sizeof(char))) == NULL )
	return( errno = ENOMEM );                /* Allocate source buffer */

    /* One byte for the number of patterns + 1 unknown + 4 for "M.K."= 6 more*/
    count1 = TITLE_L + SAMPLES * sizeof( struct mod_instr ) + PATTERNS + 6;
    if( (buf1 = (word *)calloc( (size_t)count1, sizeof(char))) == NULL )  {
	free( buf );
	return( errno = ENOMEM );                /* Allocate target buffer */
    }
    count = read( hsource, (char *)buf, count ); /* Read ST module attributes*/
    switch( (signed)count )     {
	case  0:    {   /* Unexpected EOF */
	    printf( "\"%s\" is too short to be recognized\n", source );
	    errno = EINVAL;
	case -1:        /* Read error */
	    return( release( buf, buf1 ));
	}
    }
    /*Convert and get no.of patterns used in low, unused in high word (count)*/
    if( !attrib_conv( buf, buf1, &inf ))    {
	printf( "\"%s\" is not a Scream Tracker module\n", source );
	errno = EINVAL;
	return( release( buf, buf1 ));
    }
    if( write( htarget, (char *)buf1, count1 ) == -1 )  { /* Write MOD attr.*/
	return( release( buf, buf1 ));
    }
    release( buf, buf1 );
    buf1 = NULL;		    /* To allow subsequent "release" calls */
    count  = inf.pat_used;		    /* Save number of used */
    count1 = inf.pat_to_skip;		    /* and unused patterns */

/*************************** Convert pattern data ***************************/

    if( (buf = (word *)malloc( (size_t)PAT_LEN )) == NULL )
	return( errno = ENOMEM );

    for( i = 0; i < count; i++ )        {               /* Main loop */
	ch = read( hsource, (char *)buf, PAT_LEN );     /* Read STM pattern */
	switch( ch )    {
	    case  0:    /* Unexpected EOF */
		printf( "Missing patterns in \"%s\"\n", source );
		errno = EINVAL;
	    case -1:    /* Read error */
		return( release( buf, buf1 ));
	}
	buf1 = buf;
			/* Convert channels */
	for( j = 0; j < LINES * CHANNELS; j++, buf1 += CHAN_LEN/sizeof(word) )
	    if( ( eff = channel_conv( buf1 )) != NULL )
		printf(
		"WARNING: Cannot convert %s in pattern #%d, line #%d, channel #%d\n",
		    eff, i, j / CHANNELS, j % CHANNELS + 1 );

	buf1 = NULL;    /* Don't disturb release later on */
	if( write( htarget, (char *)buf, PAT_LEN ) == -1 )
	    return( release( buf, buf1 ));              /* Write MOD pattern*/
    }

    if( count1 )        /* Skip unused patterns */
	if( lseek( hsource, (long)count1 * (long)PAT_LEN, SEEK_CUR ) == -1L )
	    return( release( buf, buf1 ));
	else
	    printf( "WARNING: %d unused patterns in \"%s\" - ignored\n",
		count1, source );

    /* Write 4 more bytes of garbage for paragraph alignment of sample data */
    if( write( htarget, (char *)buf, 4 ) == -1 )
	return( release( buf, buf1 ));
    free( buf );

/******************************* Copy samples *******************************/

    if( inf.data_len < 0xFF00L )
	count = (word)inf.data_len;	/* Allocate whole sample data area */
    else
	count = 0xFF00; 	    /* Max allowable for one read or write */

    /* Dynamically allocate a large file buffer. If there's not enough
     * memory for it, find the largest amount available on the near heap
     * and allocate that. This can't fail, no matter what the memory model.
     */
    if( (buf = (word *)malloc( (size_t)count )) == NULL )
    {
        count = _memmax();
	if( (buf = (word *)malloc( (size_t)count )) == NULL )
	    return( errno = ENOMEM );
    }

    /* Read-write until there's nothing left. */
    while( !eof( hsource ) )
    {
	if( (count = read( hsource, (char *)buf, count )) == -1 )
	    break;
	inf.data_len -= (long)count;	/* Count down bytes to copy */
	if( (count = write( htarget, (char *)buf, count )) == -1 )
	    break;
    }
    if( inf.data_len > 15 ) /* At least 1 paragraph not read - file truncated*/
	printf("ERROR: Missing %ld bytes of sample data in \"%s\" - module unusable\n",
				inf.data_len, source );
    else
	write( htarget, id, sizeof id );	  /* OK, append ID string */
    return( release( buf, buf1 ));
}

int release( word *buffer, word *buffer1 )
{
    free( buffer );     /* If buffer of buffer1 is NULL, free will ignore it*/
    free( buffer1 );
    return errno;
}
