/*
 * ToAmiga Music Module Converter Header File
 * Copyright (c) 1992 by L.Georgiev
 */

#define VER	104	/* This is the second release after 1.00 */
			/* Four changes have been made (see .c files) */

#define MAGIC   0x7800  /* A "magic" number - I don't know anything about it*/
#define TITLE_L   20    /* Amiga module title length */
#define SAMPLES   31    /* Number of samples in both kinds of modules */
#define PATTERNS 128    /* Max number of patterns in both kinds of modules */
#define CHAN_LEN   4    /* Single channel length in both kinds of modules */
#define CHANNELS   4    /* Number of channels in both kinds of modules */
#define LINES     64    /* Number of note lines in a single pattern */
#define PAT_LEN ( CHAN_LEN * CHANNELS * LINES ) /* Pattern length (1 KByte) */

/**************** Scream Tracker Command (Effect) Definitions ***************/

#define TEMPO       1
#define JUMP        2
#define BREAK       3
#define SLIDE_VOL   4
#define PORT_UP     5
#define PORT_DOWN   6
#define NOTE_PORT   7
#define VIBRATO     8
#define TREMOLO     9   /* Not implemented in Amiga modules */
#define ARPEGGIO   10
#define VOL_PORT   11   /* Not used */
#define VOL_VIBR   12   /* Not used */

#define VOL_CHNG   16   /* Not existant as a command, means a new volume */
                        /* for the current note. */

/*************************** Function Declarations **************************/

typedef unsigned word;  /* A word must always be 16-bit (not 32) */

struct hdr_info 			/* Returned by attrib_conv() */
{
    word    pat_used:8;			/* Number of patterns in sequence */
    word    pat_to_skip:8;		/* Number of patterns not used */
    long    data_len;			/* Sample data length expected */
};

int read_conv_write( char *, char * );	/*Sub-main function (calls the rest)*/
int release( word *, word * );		/* Releases memory */
					/* Converts sample attributes: */
int attrib_conv( word *, word *, struct hdr_info * );
char * channel_conv( word * );      	/* Converts a single channel */

/****************************** Data Structures ******************************

**  IMPORTANT: Scream Tracker unsigned data are aligned in DEC/Intel manner  **
**  (low byte first, then high byte), while Amiga unsigned data are aligned  **
**  in Motorola manner (high byte first, then low byte).  This affects word  **
**  operations and must be taken into account for correct results.  That is  **
**  why I have used the "swab" function ( see <stdlib.h> ) in convert.c ...  **

Alignment order in a
       MACHINE WORD:    |      Even Address      |    Next Odd address    |
	       Bit #    |15 14 13 12 11 10  9  8 | 7  6  5  4  3  2  1  0 |
Direction of alignment:  +-- <--- <--- <--- <--+   +-- <--- <--- <--- <--+
			 |                     |   |                     |
			 |    Begin here --> --+   +--> End (at last!)   |
			 |                                               |
			 +--> ---> ---> Continue this way --> ---> ---> -+
				   (watch out not to get lost!)
*/

struct  stm_header          /* Scream Tracker Module Header */
{
    char    module_name[20];    /* ASCIIZ except if 20 chars long */
    char    tracker_name[8];    /* No nul can be here */
    word    stm_sign:8;         /* Must be an EOF character (0x1A) */
    word    file_type:8;        /* 1 = song, 2 = module (with instruments) */
    word    version_major:8;
    word    version_minor:8;
    word    tempo:8;
    word    no_of_patterns:8;	/* # patterns saved (some might not be used)*/
    word    global_volume:8;
    char    reserved[12];       /* + 1 empty byte from the previous word =13*/
};

struct  stm_instr           /* Scream Tracker Module Instrument Attributes */
{
    char    name[12];           /* ASCIIZ */
    word    zero:8;             /* Must be 0 */
    word    disk:8;             /* Instrument disk (0..99) */
    word    int_seg;            /* Used as internal segment while playing */
    word    length;
    word    loop_start;
    word    loop_end;
    word    volume;
    word    mid_C_speed;        /* Playing freq.to produce C2 (Usually 8448)*/
    long    reserved;
    word    internal_seg;       /* Sample paragraph as saved */
};

struct	mod_instr	    /* Amiga Module Instrument Attributes */
{
    char    name[22];           /* ASCIIZ */
    word    length;
    word    volume;             /* 64 max */
    word    loop_start;
    word    loop_length;	/* Length, loop start and length are in words*/
};

/*                        Scream Tracker Channel Data:
			  ============================

+<-octave-->+<--note--->+<---sample--->+<-v_lo->+<-vol_hi-->+<-effect-->+<--effect_parameter--->+
|   4 bit   |   4 bit   |    5 bit     |  3 bit |   4 bit   |   4 bit   |         8 bit         |
|  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |
+<-- 1st byte, even --->+<--- 2nd byte, odd --->+<-- 3rd byte, even --->+<--- 4th byte, odd --->+
|                                               |                                               |
+<------------------ 1st word ----------------->+<------------------ 2nd word ----------------->+
*/

struct  stm_channel
{
    word    note:4;
    word    octave:4;           /* 1st word */
    word    vol_low:3;
    word    sample:5;

    word    effect:4;
    word    vol_high:4;         /* 2nd word */
    word    eff_par:8;
};

/*                        Amiga Module Channel Data:
			  ==========================
	 sample MSB
	 ^  +<------------note_data------------>+<-sample-->+<-effect-->+<--effect_parameter--->+
	 |1b|              12 bit               |   4 bit   |   4 bit   |         8 bit         |
|00|00|00|MS|  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |  |
+<-- 1st byte, even --->+<--- 2nd byte, odd --->+<-- 3rd byte, even --->+<--- 4th byte, odd --->+
|                                               |                                               |
+<------------------ 1st word ----------------->+<------------------ 2nd word ----------------->+
*/

struct  mod_channel
{
    word    note_high:4;
    word    sample_msb:4;               /* 1st word */
    word    note_data:8;

    word    effect:4;
    word    sample:4;                   /* 2nd word */
    word    eff_par:8;
};

union   channel
{
    struct stm_channel  stm;            /* This helps channel converting */
    struct mod_channel  mod;
};

/* The following is never used, but can help understanding file formats */

struct  stm_module
{
    struct      stm_header      header;
    struct      stm_instr       attrib[ SAMPLES ];
		char            order[ PATTERNS ];
    struct      stm_channel     patterns[10];
    unsigned    char            samples[ 50000 ]; /* <--For example...*/
};

struct  amiga_mod
{
		char            title[20];
    struct      mod_instr       attrib[ SAMPLES ];
		char            no_of_patterns;
		char            unknown;
		char            sequence[ PATTERNS ];
		char            M_K_[4];
    struct      mod_channel     patterns[10];
		char            samples[ 50000 ]; /* <--For example...*/
};
